# -*- coding: UTF-8 -*-
import json
from base64 import b64encode, b64decode
from pandas import DataFrame
from pyDes import PAD_PKCS5, CBC, des
from .__sqltool__.action import SqlData, Action

__Des_Key0 = '%s&@W2*<FRW2'#'%s+23das&@W2*<FRW2'
__Des_IV = b'\x52\x63\x78\x61\xBC\x48\x6A\x07'


def encrypt(user, passwd):
    k = des((__Des_Key0 % user)[:8], CBC, __Des_IV, padmode=PAD_PKCS5)
    return b64encode(k.encrypt(passwd)).decode('utf-8')


def decrypt(user, passwd_enc):
    k = des((__Des_Key0 % user)[:8], CBC, __Des_IV, padmode=PAD_PKCS5)
    return k.decrypt(b64decode(passwd_enc)).decode('utf-8')


# 表格去重
def table_drop_duplicates(dts: list, *lies) -> DataFrame:
    df = DataFrame(data=dts)
    df.drop_duplicates(subset=lies, keep='first', inplace=True)
    return df


# 数据保存工具
class DataUpdateSave:
    def __init__(self, sqlt: Action, main_keys: list, iftz=True):
        self.sqlt = sqlt
        self.main_keys = main_keys
        self.iftz = iftz

    def getWhere(self, datadts):
        assert datadts
        wheres = []
        for dt in datadts:
            txt = "','".join([str(dt[k]) for k in self.main_keys])
            wheres.append(f"('{txt}')")
        # 删除主表主键数据
        where = f"({','.join(self.sqlt.getLiestrs(self.main_keys))}) in ({','.join(wheres)})"
        return where

    def getChildDts(self, datadts, chlie, if_pop=True) -> list:
        all_chls = []
        for dt in datadts:
            if if_pop:
                chls = dt.pop(chlie, [])
            else:
                chls = dt.get(chlie, [])
            # 自动转换
            if type(chls) == str: chls = json.loads(chls)
            for key in self.main_keys:
                for chdt in chls:
                    chdt[key] = dt[key]
            all_chls.extend(chls)
        return all_chls

    def updateMain(self, datadts: list, table, json_lies: list = (), colmap: dict = {}):
        if not datadts:
            if self.iftz: print('没有主数据')
            return False
        where = self.getWhere(datadts)
        if self.sqlt.ifExist(table):
            if self.iftz: print('删除主表主键')
            self.sqlt.delete(table, where)
        else:
            if self.iftz: print('暂无主表:', table)
        for lie in json_lies:
            colmap[lie] = 'text'
            for dt in datadts:
                dt[lie] = json.dumps(dt.get(lie, ''), ensure_ascii=False)
        # 插入数据
        sqldata = SqlData(datadts)
        self.sqlt.createTable(table, sqldata.getLies(), colmap=colmap, key=self.main_keys)
        self.sqlt.insert(table, sqldata)
        if self.iftz: print(table, '主表数据量:', len(sqldata))

    def updateChild(self, datadts, child_table, chlie, if_pop=True, colmap=None):
        chdts = self.getChildDts(datadts, chlie, if_pop=if_pop)
        if chdts:
            where = self.getWhere(datadts)
            if self.sqlt.ifExist(child_table):
                if self.iftz: print('删除子表主键')
                self.sqlt.delete(child_table, where)
            else:
                if self.iftz: print('暂无子表:', child_table)
            # 插入数据
            sqldata = SqlData(datadts)
            self.sqlt.createTable(child_table, sqldata.getLies(), colmap=colmap)
            self.sqlt.insert(child_table, *chdts)
            if self.iftz: print(child_table, '子表数据量:', len(sqldata))
        else:
            if self.iftz: print('没有子表数据!')

    def commit(self):
        self.sqlt.commit()
        if self.iftz: print('提交成功!')



import os
import sqlite3
from .__sqltool__.action import Action, SqlData
from .tools import decrypt
from .__sqltool__.alter import Alter


class Sqlite(Action, Alter):
    def __init__(self, dbfilepath, if_new=True, ifencryp=False):
        self._dbfilepath = dbfilepath
        self._if_new = if_new
        self._ifencryp = ifencryp
        class_map = {'varchar': 'text',
                     'text': 'text',
                     'bool': 'integer',
                     'int': 'integer',
                     'float': 'real'
                     }
        super().__init__(self._db_func, placeholder='?', class_map=class_map)

    # 打开数据库连接
    def db_func(self):
        if not self._if_new and not os.path.exists(self._dbfilepath):
            raise ValueError(f'没有 {self._dbfilepath} 文件!')
        db = sqlite3.connect(self._dbfilepath)
        return db

    def getTablestr(self, table, **kwargs):
        if '.' in table:
            t1, t2 = table.split('.')
            return f'"{t1}"."{t2}"'
        else:
            return f'"{table}"'

    def getLiestr(self, lie, **kwargs):
        if lie in {'*', '1', 'count(1)'}:
            return lie
        else:
            return f'"{lie}"'

    # 判断表是否存在
    def ifExist(self, table):
        b = self.run(f"SELECT 1 FROM sqlite_master WHERE type='table' AND name='{table}'")
        return b and len(self._cursor.fetchall()) > 0



# from pymssql import connect as sst_connect
from .__sqltool__.action import Action, SqlData
from .tools import decrypt
from .__sqltool__.alter import Alter
import pyodbc


class Sqlserver(Action, Alter):
    def __init__(self, dbname, user, pwd,
                 host='127.0.0.1', port: int = 1433,
                 charset='UTF8',  # cp936
                 ifencryp=True):
        self._dbname = dbname
        self._user = user
        self._pwd = pwd
        self._host = host
        self._port = port
        self._charset = charset
        self._ifencryp = ifencryp
        class_map = {'varchar': 'nvarchar(255)',
                     'text': 'ntext',
                     'bool': 'bit',
                     'int': 'int',
                     'float': 'float'
                     }
        super().__init__(self._db_func, placeholder='?', class_map=class_map)

    def _db_func(self):
        db = pyodbc.connect(
            f'DRIVER={{ODBC Driver 17 for SQL Server}};SERVER={self._host};DATABASE={self._dbname};UID={self._user};PWD={decrypt(self._user, self._pwd) if self._ifencryp else self._pwd}')  # DRIVER={SQL Server};
        # db = sst_connect(host=self.host, port=self.port,
        #                      user=self.user, password=self.passwd,
        #                      database=self.dbname, charset=self.charset)
        return db

    def getTablestr(self, table, **kwargs):
        if '.' in table:
            t1, t2 = table.split('.')
            return f'[{t1}].[{t2}]'
        else:
            return f'[{table}]'

    def getLiestr(self, lie, **kwargs):
        if lie in {'*', '1', 'count(1)', 'top 1 1'}:
            return lie
        else:
            return f"[{lie}]"

    def deleteTable(self, table):
        # DROP TABLE table_name
        sql = '''
            If Exists (select * from sysobjects where id = object_id('{table}') and OBJECTPROPERTY(id, 'IsUserTable') = 1)
                DROP TABLE {table} 
        '''.format(table=self.getTablestr(table))
        return self.run(sql)

    # 判断表是否存在
    def ifExist(self, table):
        return Action.ifExist(self, table, def_schema='dbo')



import traceback
from pandas import DataFrame, isna


class Body:
    def __init__(self, db_func, placeholder, class_map: dict):
        self._db_func = db_func
        self._db = db_func()
        self._cursor = self._db.cursor()

        self.placeholder = placeholder
        self.class_map = class_map

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def commit(self):
        self._db.commit()

    def close(self):
        try:
            self._cursor.close()
            self._db.close()
        except:
            # traceback.print_exc()
            pass

    # 刷新连接
    def refresh(self):
        self.close()
        self._db = self._db_func()
        self._cursor = self._db.cursor()

    # 事务回滚
    def rollback(self):
        self._db.rollback()

    # 使用sql操作
    def run(self, sql):
        # print(sql)
        return self._cursor.execute(sql)

    # 获取结果
    def getResult(self, data_class='dts'):
        # 获取数据
        lies = [lc[0] for lc in self._cursor.description] if self._cursor.description else []
        result = list()
        for row in self._cursor.fetchall():
            result.append({lie: data for lie, data in zip(lies, row)})
        # 处理返回格式
        if data_class == 'ls':
            result = [list(dt.values()) for dt in result]
        elif data_class == 'df':
            result = DataFrame(data=result)
        return lies, result

    # 判断表是否存在
    def ifExist(self, table, def_schema):
        ls = table.split('.')
        if len(ls) == 1:
            schema = def_schema
        else:
            schema, table = ls
        b = self.run(
            f"SELECT 1 FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = '{table}' and table_schema='{schema}'")
        return b and len(self._cursor.fetchall()) > 0

    def in_run(self, sql, lies, *datas, if_error_one=True):
        try:
            self._cursor.executemany(sql, datas)
        except Exception as e:
            self.rollback()
            if if_error_one:
                print("\033[0;31m", '批量插入出错,执行单条插入', "\033[0m")
                for i, data in enumerate(datas):
                    try:
                        self._cursor.executemany(sql, [data])
                    except:
                        print("\033[0;32m", sql, "\033[0m")
                        traceback.print_exc()
                        print("\033[0;31m", f'具体数据出错-index{i}', "\033[0m")
                        [print("\033[0;31m", lie, ':', v, "\033[0m") for lie, v in zip(lies, data)]
                        self.rollback()
                        break
            raise e

    def getTablestr(self, table, **kwargs) -> str:
        raise TypeError('需要实现方法-getTablestr')

    def getLiestr(self, lie, **kwargs) -> str:
        raise TypeError('需要实现方法-getLiestr')

    def getLiestrs(self, lies, **kwargs) -> list:
        if type(lies) == str: lies = [lies]
        return [self.getLiestr(lie, **kwargs) for lie in lies]




from .base.body import Body, traceback
from .base.sql_data import SqlData


class Action(Body):
    def createTable(self, table, lies: list,
                    colmap: dict = None,
                    key: list or str = None,
                    hz: str = None, **null):
        colmap = colmap if colmap else {}
        assert lies, "数量有误！"
        if key:
            if type(key) == str:
                key = [key]
                key = f" ,PRIMARY KEY({','.join(self.getLiestrs(key))})"
        else:
            key = ''
        # 默认类型为短字符串类型
        liestr = ",\n".join(f'{self.getLiestr(lie)} {colmap.get(lie, self.class_map["varchar"])}' for lie in lies)
        sql = '''
            Create Table If Not Exists {table}
            ({lies} {key}){hz}
        '''.format(table=self.getTablestr(table), lies=liestr, key=key, hz=hz if hz else '')
        return self.run(sql)

    # 删除表
    def deleteTable(self, table):
        # DROP TABLE table_name
        sql = '''DROP TABLE IF EXISTS {table} 
        '''.format(table=self.getTablestr(table))
        return self.run(sql)

    def _auto_add_lie(self, lies, table, dataclass, other):
        have_lies = self.select('*', table, where='1=0', ifgetlies=True)[0]
        not_liest = set(lies) - set(have_lies)
        sql = '''ALTER TABLE {table} {add} {other}
                '''.format(table=self.getTablestr(table),
                           add=','.join(f'ADD {lie} {dataclass}' for lie in self.getLiestrs(not_liest)),
                           other=other)
        return self.run(sql)

    # 插入
    def insert(self, table: str, sql_data: SqlData,
               lies=None,
               in_hz='',
               auto_lie_class=None,
               if_error_one=True, **null):
        assert sql_data, "数量不能为0！"
        lies = [lie.strip() for lie in lies] if lies else sql_data.getLies()
        datas = sql_data.getValues(lies)
        # 插入语句
        sql = '''INSERT INTO {table}({liestr})
                VALUES ({cstr}) {in_hz}
        '''.format(table=self.getTablestr(table), liestr=','.join(self.getLiestrs(lies)),
                   cstr=','.join([self.placeholder] * len(lies)), in_hz=in_hz)
        try:
            return self.in_run(sql, lies, *datas, if_error_one=if_error_one)
        except Exception as e:
            if auto_lie_class:
                self._auto_add_lie(lies, table, auto_lie_class, other='')
                return self.in_run(sql, lies, *datas, if_error_one=if_error_one)
            else:
                raise e

    def create_insert(self, table, dts: list,
                      colmap: dict = {},
                      key: list or str = None,
                      hz: str = None,
                      in_hz='',
                      auto_lie_class=None, **null):
        sdata = SqlData(dts)
        self.createTable(table, sdata.getLies(), sdata.getAutoColMap(self.class_map, **colmap), key=key, hz=hz, **null)
        self.insert(table, sdata, in_hz=in_hz, auto_lie_class=auto_lie_class, **null)

    # 修改
    def update(self, table, dt: dict, where: str, **kwargs):
        assert dt, "数量有误！"
        sdata = SqlData([dt])
        lies = sdata.getLies()
        setv = ','.join([f"{lie}={self.placeholder}" for lie in self.getLiestrs(lies)])
        sql = '''UPDATE {table}
                SET {setv}
                WHERE {where}
        '''.format(table=self.getTablestr(table, **kwargs), setv=setv, where=where)
        # print(sql)
        return self.in_run(sql, lies, *sdata.getValues(lies), **kwargs)

    # 删除
    def delete(self, table, where: str):
        # 表名
        sql = '''DELETE FROM {table}
            WHERE {where}
        '''.format(table=self.getTablestr(table), where=where)
        return self.run(sql)

    def select(self, lies, table, where=None, other='',
               data_class='dts',
               ifgetlies=False,
               ifget_one_lie=False,
               if_original_table=False,
               if_original_lies=False) -> list or (list, list):
        assert not (if_original_lies and type(lies) == str), '开启原始列时,输入列必须为列表格式!'
        sql = '''select {lies}
                   from {table}
                   where {where}
                   {other}
           '''.format(lies=','.join(lies if if_original_lies else self.getLiestrs(lies)),
                      table=table if if_original_table else self.getTablestr(table),
                      where=where if where else '1=1',
                      other=other)
        self.run(sql)
        # 获取数据
        lies, result = self.getResult(data_class='ls' if ifget_one_lie else data_class)
        if ifget_one_lie:
            result = [d[0] for d in result]
        if ifgetlies:
            return lies, result
        else:
            return result

    # 判断是否可以查询到
    def ifGet(self, table, where=None, if_error=True):
        try:
            if len(self.select('1', table, where=where, data_class='ls', other='limit 1')) > 0:
                return True
            else:
                return False
        except:
            if if_error:
                traceback.print_exc()
            return False

    # 获取数量
    def getNum(self, table, where=None):
        num = self.select('count(1)', table, where=where, data_class='ls')[0][0]
        return num



from pymysql import connect as myt_connect
from .__sqltool__.action import Action, SqlData
from .tools import decrypt
from .__sqltool__.alter import Alter


class Mysql(Action, Alter):
    def __init__(self, dbname, user, pwd,
                 host='127.0.0.1', port: int = 3306,
                 charset='utf8mb4',
                 ifencryp=True):
        self._dbname = dbname
        self._user = user
        self._pwd = pwd
        self._host = host
        self._port = port
        self._charset = charset
        self._ifencryp = ifencryp
        class_map = {'varchar': 'varchar(255)',
                     'text': 'text',
                     'bool': 'blob',
                     'int': 'int',
                     'float': 'float'
                     }
        super().__init__(self._db_func, placeholder='%s', class_map=class_map)

    def _db_func(self):
        db = myt_connect(host=self._host, port=self._port,
                         user=self._user, passwd=decrypt(self._user, self._pwd) if self._ifencryp else self._pwd,
                         database=self._dbname, charset=self._charset)
        return db

    def run(self, sql):
        # 设置重连
        self._db.ping(reconnect=True)
        return Action.run(self, sql)

    def in_run(self, sql, *datas, if_error_one=True):
        # 设置重连
        self._db.ping(reconnect=True)
        return Action.in_run(self, sql, *datas, if_error_one=if_error_one)

    def commit(self):
        # 设置重连
        self._db.ping(reconnect=True)
        return Action.commit(self)

    def getTablestr(self, table, **kwargs):
        return '`%s`' % table

    def getLiestr(self, lie, **kwargs):
        if lie in {'*', '1', 'count(1)'}:
            return lie
        else:
            return f"`{lie.replace('%', '%%')}`"

    def ifExist(self, table):
        b = self.run(f'show tables like "{table}"')
        return b and len(self._cursor.fetchall()) > 0



from pandas import DataFrame
import json


def _make_func():
    def temp(data):
        if type(data) == str:
            data = data.replace("'", "\\'").strip()
        if type(data) in (tuple, list, dict):
            data = json.dumps(data, ensure_ascii=False)
        return data

    return temp


# 用于操作的集成数据库类型
class SqlData:
    def __init__(self, data: list[dict] or DataFrame):
        df = data if type(data) == DataFrame else DataFrame(data=data)
        df.rename(columns={lie: lie.strip() for lie in df.columns}, inplace=True)
        self._data = df.where(df.notnull(), None).applymap(_make_func())

    def __bool__(self):
        return len(self._data) > 0

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return self._data

    # 获取
    def getLies(self) -> list:
        return list(self._data.columns)

    def getDf(self) -> DataFrame:
        return self._data

    def getDts(self) -> list[dict]:
        return self._data.to_dict(orient='records')

    def getValues(self, lies: list = None) -> list[tuple]:
        if lies:
            return list(zip(*[self._data[lie].tolist() for lie in lies]))
        else:
            return self._data.to_records(index=False).tolist()

    def getAutoColMap(self, class_map: dict, **colmap) -> dict:
        dt = {}
        # 以首行数据为基准进行自动类型分配
        for lie, value in self._data.iloc[0].to_dict().items():
            if colmap.get(lie):
                dt[lie] = colmap[lie]
                continue
            if type(value) == str:
                if len(lie) < 255:
                    dt[lie] = class_map['varchar']
                else:
                    dt[lie] = class_map['text']
            elif type(value) == bool:
                dt[lie] = class_map['bool']
            elif type(value) == int:
                dt[lie] = class_map['int']
            elif type(value) == float:
                dt[lie] = class_map['float']
            else:
                dt[lie] = class_map['varchar']
        return dt
